@@ -12,50 +12,56 @@ Author: Ye Chang
 Date: 2025-01-27
 """
 
+import logging
 import os
 import tempfile
 import threading
 import time
 from collections import defaultdict
+from collections.abc import Iterator
 from concurrent.futures import ProcessPoolExecutor, as_completed
-from typing import Any, Dict, Iterator, List
-import logging
+from typing import Any
 
 import pysam
 from rich.progress import Progress, TimeElapsedColumn, TimeRemainingColumn
 
+from .utils import (
+    Fragment,
+    calculate_average_base_quality,
+    cluster_umis_by_edit_distance_frequency_aware,
+    extract_umi,
+    get_read_position,
+)
+
 # Set up logger
 logger = logging.getLogger(__name__)
 if not logger.handlers:
     handler = logging.StreamHandler()
-    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
+    formatter = logging.Formatter(
+        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
+    )
     handler.setFormatter(formatter)
     logger.addHandler(handler)
     logger.setLevel(logging.WARNING)  # Only show warnings and errors by default
 
-# Import utility functions
-from .utils import (
-    Fragment,
-    calculate_average_base_quality,
-    cluster_umis_by_edit_distance_frequency_aware,
-    get_read_position,
-    extract_umi,
-)
-
 # Globals used inside worker processes (set by pool initializer)
 WORKER_READER = None
 WORKER_WRITER = None
 WORKER_SHARD_PATH = None
 
+
 def _worker_initializer(bam_path: str, shard_dir: str):
     """Initializer for ProcessPool workers: open per-process reader and writer.
 
     Each process opens one AlignmentFile for reading and one shard BAM for writing.
     """
-    import os as _os
     import atexit as _atexit
+    import os as _os
+
     import pysam as _pysam
+
     global WORKER_READER, WORKER_WRITER, WORKER_SHARD_PATH
+    # Only create reader once per worker process
     if WORKER_READER is None:
         WORKER_READER = _pysam.AlignmentFile(bam_path, "rb")
     pid = _os.getpid()
@@ -101,14 +107,15 @@ def _worker_shutdown_task():
 # QUALITY SCORING FUNCTIONS
 # ============================================================================
 
+
 def calculate_fragment_quality_score(fragment, best_read_by):
     """
     Calculate quality score for a fragment based on the specified criteria.
-    
+
     Args:
         fragment: Fragment object containing read(s)
         best_read_by: Quality criterion ('mapq' or 'avg_base_q')
-        
+
     Returns:
         float: Quality score for the fragment
     """
@@ -146,10 +153,11 @@ def calculate_fragment_quality_score(fragment, best_read_by):
 # BAM READER CLASS
 # ============================================================================
 
+
 class BAMReader:
     """
     Unified BAM file reader that handles genomic window creation.
-    
+
     This class provides a unified interface for reading BAM files and creating
     genomic windows for parallel processing. It supports different reading
     strategies and efficiently checks for reads in genomic regions.
@@ -187,12 +195,8 @@ class BAMReader:
             # Handle pysam internal errors or file access issues silently
             return False
 
-    def get_windows(self) -> Iterator[Dict[str, Any]]:
-        """Get all windows for processing using sequential approach."""
-        yield from self._create_sequential_windows()
-
-    def _create_sequential_windows(self) -> Iterator[Dict[str, Any]]:
-        """Create genomic windows for sequential processing with proper overlapping."""
+    def get_windows(self) -> Iterator[dict[str, Any]]:
+        """Create and yield genomic windows for processing."""
         import time
 
         window_id = 0
@@ -223,15 +227,12 @@ class BAMReader:
                 window_id += 1
                 total_windows_created += 1
             else:
-                # Long chromosome - split into overlapping windows
-                # Step by window_size, with search region extended by max_pair_dist
+                # Long chromosome - split into windows
                 for i in range(0, contig_len, self.window_size):
                     window_start = i
                     window_end = min(i + self.window_size, contig_len)
                     search_start = max(0, window_start)
-                    search_end = min(
-                        contig_len, window_end + self.max_pair_dist
-                    )
+                    search_end = min(contig_len, window_end + self.max_pair_dist)
 
                     # Determine if this is the first window
                     is_first_window = i == 0
@@ -250,7 +251,9 @@ class BAMReader:
 
         # Print completion with timing
         window_creation_time = time.time() - start_time
-        print(f"✅ Created {total_windows_created} windows (⏱️ {window_creation_time:.2f}s)")
+        print(
+            f"✅ Created {total_windows_created} windows (⏱️ {window_creation_time:.2f}s)"
+        )
 
     def _create_window_data(
         self,
@@ -261,7 +264,7 @@ class BAMReader:
         window_end: int,
         window_id: int,
         is_first: bool,
-    ) -> Dict[str, Any]:
+    ) -> dict[str, Any]:
         """Create window data dictionary."""
         return {
             "window_id": f"{chromosome}_window_{window_id}",
@@ -280,10 +283,11 @@ class BAMReader:
 # THREAD-SAFE WRITER CLASS
 # ============================================================================
 
+
 class ThreadSafeWriter:
     """
     Thread-safe BAM writer using multiple temporary files.
-    
+
     This class provides thread-safe writing capabilities by using separate
     temporary files for each thread, then merging them at the end. It uses
     persistent file handles for efficiency."""
@@ -324,7 +328,7 @@ class ThreadSafeWriter:
                 temp_file, "wb", header=self.header
             )
 
-    def write_reads(self, reads: List[str], header: Any, thread_id: int = None):
+    def write_reads(self, reads: list[str], header: Any, thread_id: int = None):
         """Write reads to the temp file for the specified thread using persistent file handle."""
         if not reads:
             return
@@ -375,15 +379,19 @@ class ThreadSafeWriter:
                 print(f"🧹 Cleaned up temporary directory: {self.temp_dir}")
             except Exception as e:
                 logger.warning(f"Could not remove temp directory {self.temp_dir}: {e}")
-                print(f"⚠️  Warning: Could not remove temp directory {self.temp_dir}: {e}")
+                print(
+                    f"⚠️  Warning: Could not remove temp directory {self.temp_dir}: {e}"
+                )
                 # Try to remove individual files as fallback
                 try:
                     for temp_file in self.temp_files.values():
                         if os.path.exists(temp_file):
                             os.unlink(temp_file)
                 except Exception as cleanup_error:
-                    logger.warning(f"Failed to cleanup individual temp files: {cleanup_error}")
-        
+                    logger.warning(
+                        f"Failed to cleanup individual temp files: {cleanup_error}"
+                    )
+
         # Clear references to help with garbage collection
         self.temp_files.clear()
         self.file_handles.clear()
@@ -394,13 +402,14 @@ class ThreadSafeWriter:
 # WINDOW PROCESSING FUNCTIONS
 # ============================================================================
 
-def process_window(window_data: Dict[str, Any]) -> Dict[str, Any]:
+
+def process_window(window_data: dict[str, Any]) -> dict[str, Any]:
     """
     Process a single genomic window for deduplication.
-    
+
     Args:
         window_data: Dictionary containing window parameters and configuration
-        
+
     Returns:
         Dict containing processing results and statistics
     """
@@ -458,8 +467,14 @@ def process_window(window_data: Dict[str, Any]) -> Dict[str, Any]:
         if method == "coordinate":
             # Coordinate-based deduplication
             deduplicated_reads, read_stats = deduplicate_reads_by_coordinate(
-                reads, max_pair_dist, start_only, end_only, best_read_by, keep_duplicates,
-                fragment_paired, fragment_mapped
+                reads,
+                max_pair_dist,
+                start_only,
+                end_only,
+                best_read_by,
+                keep_duplicates,
+                fragment_paired,
+                fragment_mapped,
             )
         else:
             # UMI-based deduplication
@@ -482,38 +497,32 @@ def process_window(window_data: Dict[str, Any]) -> Dict[str, Any]:
         global WORKER_WRITER
         # Lazily create writer only when we have output
         if deduplicated_reads and WORKER_WRITER is None:
-            WORKER_WRITER = pysam.AlignmentFile(WORKER_SHARD_PATH, "wb", header=WORKER_READER.header)
+            WORKER_WRITER = pysam.AlignmentFile(
+                WORKER_SHARD_PATH, "wb", header=WORKER_READER.header
+            )
 
         for r in deduplicated_reads:
-            # Ensure reference_id is set when writing
-            if r.reference_id is None and r.reference_name is not None:
-                try:
-                    r.reference_id = WORKER_READER.get_tid(r.reference_name)
-                except Exception:
-                    pass
             WORKER_WRITER.write(r)
-        # Encourage timely BGZF block flush
-        if deduplicated_reads:
-            try:
-                WORKER_WRITER.flush()
-            except Exception:
-                pass
 
         # Calculate statistics
         original_count = len(reads)
         deduplicated_count = len(deduplicated_reads)
         duplicates_removed = original_count - deduplicated_count
-        
+
         # Calculate duplicates detected (for keep_duplicates mode)
         duplicates_detected = 0
         if keep_duplicates:
             # Count reads marked as duplicates
-            duplicates_detected = sum(1 for read in deduplicated_reads if read.is_duplicate)
-        
+            duplicates_detected = sum(
+                1 for read in deduplicated_reads if read.is_duplicate
+            )
+
         # Calculate appropriate rate based on keep_duplicates setting
         if keep_duplicates:
             deduplication_rate = (
-                (duplicates_detected / original_count * 100) if original_count > 0 else 0
+                (duplicates_detected / original_count * 100)
+                if original_count > 0
+                else 0
             )
         else:
             deduplication_rate = (
@@ -535,7 +544,9 @@ def process_window(window_data: Dict[str, Any]) -> Dict[str, Any]:
         }
 
     except Exception as e:
-        logger.error(f"Error processing window {window_data.get('window_id', 'unknown')}: {e}")
+        logger.error(
+            f"Error processing window {window_data.get('window_id', 'unknown')}: {e}"
+        )
         return {
             "window_id": window_data.get("window_id", "unknown"),
             "error": str(e),
@@ -547,18 +558,15 @@ def process_window(window_data: Dict[str, Any]) -> Dict[str, Any]:
         }
 
 
-
-
-
-
 # ============================================================================
 # UNIFIED PROCESSOR CLASS
 # ============================================================================
 
+
 class UnifiedProcessor:
     """
     Unified processor for all deduplication approaches and methods.
-    
+
     This class provides a unified interface for processing BAM files with
     different deduplication methods (UMI-based and coordinate-based) and
     various configuration options. It handles parallel processing, window
@@ -677,7 +685,8 @@ class UnifiedProcessor:
             # Clean, organized output
             console.print()
             # Human-readable duration formatter
-            duration_seconds = self.stats['processing_time']
+            duration_seconds = self.stats["processing_time"]
+
             def _format_duration(sec: float) -> str:
                 if sec < 60:
                     return f"{sec:.2f}s"
@@ -709,11 +718,12 @@ class UnifiedProcessor:
             stats_table.add_row(
                 "Total reads processed:", f"{self.stats['total_reads_processed']:,}"
             )
-            
+
             # Show appropriate duplicate statistic based on keep_duplicates setting
             if self.keep_duplicates:
                 stats_table.add_row(
-                    "Duplicates detected:", f"{self.stats['total_duplicates_detected']:,}"
+                    "Duplicates detected:",
+                    f"{self.stats['total_duplicates_detected']:,}",
                 )
             else:
                 stats_table.add_row(
@@ -758,7 +768,7 @@ class UnifiedProcessor:
             # Cleanup is now handled before statistics display
             pass
 
-    def _process_windows_parallel(self, windows: List[Dict[str, Any]]):
+    def _process_windows_parallel(self, windows: list[dict[str, Any]]):
         """Process windows in parallel with progress tracking."""
         import time
 
@@ -784,7 +794,11 @@ class UnifiedProcessor:
                 total_windows=total_windows,
             )
 
-            with ProcessPoolExecutor(max_workers=self.max_processes, initializer=_worker_initializer, initargs=(self.bam_file, self.shard_dir)) as executor:
+            with ProcessPoolExecutor(
+                max_workers=self.max_processes,
+                initializer=_worker_initializer,
+                initargs=(self.bam_file, self.shard_dir),
+            ) as executor:
                 # Submit all windows for processing
                 future_to_window = {}
                 for i, window in enumerate(windows):
@@ -833,7 +847,7 @@ class UnifiedProcessor:
                     # Count skipped windows
                     if not result.get("has_reads", False):
                         total_windows_skipped += 1
-                    
+
                     # Count duplicates detected
                     total_duplicates_detected += result.get("duplicates_detected", 0)
 
@@ -893,7 +907,7 @@ class UnifiedProcessor:
                 shutdown_futures = []
                 for _ in range(self.max_processes):
                     shutdown_futures.append(executor.submit(_worker_shutdown_task))
-                
+
                 # Wait for all shutdowns to complete
                 for future in as_completed(shutdown_futures):
                     try:
@@ -914,10 +928,17 @@ class UnifiedProcessor:
         temp_files = []
         if os.path.isdir(self.shard_dir):
             for name in os.listdir(self.shard_dir):
-                if name.endswith('.bam'):
+                if name.endswith(".bam"):
                     temp_files.append(os.path.join(self.shard_dir, name))
         if not temp_files:
-            print("⚠️  No temp files to process")
+            print("⚠️  No temp files to process - creating empty output")
+            # Create an empty BAM file with header only
+            with pysam.AlignmentFile(self.bam_file, "rb") as _hfsrc:
+                header = _hfsrc.header
+            with pysam.AlignmentFile(
+                self.output_file, "wb", header=header
+            ) as empty_bam:
+                pass  # Write header only
             return
 
         with Progress(
@@ -959,27 +980,44 @@ class UnifiedProcessor:
                 while time.time() < deadline:
                     valid_temp_files = []
                     for temp_file in temp_files:
-                        if os.path.exists(temp_file) and os.path.getsize(temp_file) > 0 and _is_valid_bam(temp_file):
+                        if (
+                            os.path.exists(temp_file)
+                            and os.path.getsize(temp_file) > 0
+                            and _is_valid_bam(temp_file)
+                        ):
                             valid_temp_files.append(temp_file)
                     # If we have at least one valid shard and no invalid ones remain, proceed
-                    if len(valid_temp_files) == len([p for p in temp_files if os.path.exists(p) and os.path.getsize(p) > 0]):
+                    if len(valid_temp_files) == len(
+                        [
+                            p
+                            for p in temp_files
+                            if os.path.exists(p) and os.path.getsize(p) > 0
+                        ]
+                    ):
                         break
                     time.sleep(0.1)
 
                 if not valid_temp_files:
-                    print("⚠️  No valid temp files found after waiting - creating empty output")
+                    print(
+                        "⚠️  No valid temp files found after waiting - creating empty output"
+                    )
                     # Create an empty BAM file with header only
                     with pysam.AlignmentFile(self.bam_file, "rb") as _hfsrc:
                         header = _hfsrc.header
-                    with pysam.AlignmentFile(merged_temp_file, "wb", header=header) as empty_bam:
+                    with pysam.AlignmentFile(
+                        merged_temp_file, "wb", header=header
+                    ) as empty_bam:
                         pass  # Write header only
                 elif len(valid_temp_files) == 1:
                     # For single file, just copy it
                     import shutil
+
                     shutil.copy2(valid_temp_files[0], merged_temp_file)
                 else:
                     # Use pysam.cat for efficient concatenation
-                    pysam.cat("-o", merged_temp_file, *valid_temp_files, catch_stdout=False)
+                    pysam.cat(
+                        "-o", merged_temp_file, *valid_temp_files, catch_stdout=False
+                    )
 
                 # Update progress for sorting
                 progress.update(
@@ -988,10 +1026,14 @@ class UnifiedProcessor:
                     description=f"🔀 Sorting BAM file with {self.max_processes} threads...",
                 )
 
-                # Sort the merged temp file using pysam.sort
+                # Sort the merged temp file using pysam.sort with memory optimization
+                # Use 1GB per thread for faster sorting without excessive memory pressure
+                memory_per_thread = "1G"
                 pysam.sort(
                     "-@",
                     str(self.max_processes),
+                    "-m",
+                    memory_per_thread,
                     "-o",
                     self.output_file,
                     merged_temp_file,
@@ -1024,6 +1066,7 @@ class UnifiedProcessor:
                 # Cleanup shard dir
                 try:
                     import shutil as _shutil
+
                     _shutil.rmtree(self.shard_dir)
                 except Exception:
                     pass
@@ -1068,7 +1111,6 @@ class UnifiedProcessor:
                 )
 
 
-
 def find_mate(read, reads_with_same_name, max_pair_dist):
     """Find the mate of a paired-end read using name-based lookup."""
     for candidate in reads_with_same_name:
@@ -1085,6 +1127,7 @@ def find_mate(read, reads_with_same_name, max_pair_dist):
 # DEDUPLICATION FUNCTIONS
 # ============================================================================
 
+
 def deduplicate_reads_by_umi(
     reads,
     max_pair_dist=2000,
@@ -1101,11 +1144,11 @@ def deduplicate_reads_by_umi(
 ):
     """
     Deduplicate reads using UMI-based clustering with edit distance.
-    
+
     This function groups reads by UMI and fragment position, then clusters
     similar UMIs using edit distance to identify duplicates. Uses frequency-aware
     clustering to prevent unrealistic merging of high-frequency UMIs.
-    
+
     Args:
         reads: List of pysam.AlignedSegment objects
         max_pair_dist: Maximum distance for paired-end reads
@@ -1117,7 +1160,7 @@ def deduplicate_reads_by_umi(
         keep_duplicates: Whether to keep duplicate reads and mark them
         fragment_paired: Keep only fragments with both reads present
         fragment_mapped: Keep only fragments where both reads are mapped
-        
+
     Returns:
         Tuple of (deduplicated_reads, read_stats)
     """
@@ -1231,7 +1274,9 @@ def deduplicate_reads_by_umi(
                 position_key = (frag_start, is_reverse)
             else:
                 # For single-end reads: use biological start position and strand
-                bio_start = fragment.get_biological_position(start_only=True, end_only=False)
+                bio_start = fragment.get_biological_position(
+                    start_only=True, end_only=False
+                )
                 position_key = (bio_start, fragment.read1.is_reverse)
 
         fragments_by_position[position_key].append(fragment)
@@ -1247,8 +1292,10 @@ def deduplicate_reads_by_umi(
                     fragments_by_umi[fragment.umi].append(fragment)
 
                 # Apply frequency-aware edit distance clustering
-                clustered_fragment_groups = cluster_umis_by_edit_distance_frequency_aware(
-                    fragments_by_umi, min_edit_dist_frac, min_frequency_ratio
+                clustered_fragment_groups = (
+                    cluster_umis_by_edit_distance_frequency_aware(
+                        fragments_by_umi, min_edit_dist_frac, min_frequency_ratio
+                    )
                 )
             else:
                 # No edit distance clustering - group by exact UMI
@@ -1281,8 +1328,10 @@ def deduplicate_reads_by_umi(
                         fragment_group,
                         key=lambda f: calculate_fragment_quality_score(f, best_read_by),
                     )
-                    cluster_name = best_fragment.get_cluster_name("umi", start_only, end_only)
-                    
+                    cluster_name = best_fragment.get_cluster_name(
+                        "umi", start_only, end_only
+                    )
+
                     for fragment in fragment_group:
                         is_best = fragment == best_fragment
                         fragment.read1.set_tag("cn", cluster_name)
@@ -1290,7 +1339,7 @@ def deduplicate_reads_by_umi(
                         if not is_best:
                             fragment.read1.is_duplicate = True
                         deduplicated_reads.append(fragment.read1)
-                        
+
                         if fragment.read2:
                             fragment.read2.set_tag("cn", cluster_name)
                             fragment.read2.set_tag("cs", len(fragment_group))
@@ -1303,7 +1352,9 @@ def deduplicate_reads_by_umi(
                         fragment_group,
                         key=lambda f: calculate_fragment_quality_score(f, best_read_by),
                     )
-                    cluster_name = best_fragment.get_cluster_name("umi", start_only, end_only)
+                    cluster_name = best_fragment.get_cluster_name(
+                        "umi", start_only, end_only
+                    )
                     best_fragment.read1.set_tag("cn", cluster_name)
                     best_fragment.read1.set_tag("cs", len(fragment_group))
                     deduplicated_reads.append(best_fragment.read1)
@@ -1333,10 +1384,10 @@ def deduplicate_reads_by_coordinate(
 ):
     """
     Deduplicate reads using coordinate-based clustering.
-    
+
     This function groups reads by fragment position and selects the best
     read from each group based on quality criteria.
-    
+
     Args:
         reads: List of pysam.AlignedSegment objects
         max_pair_dist: Maximum distance for paired-end reads
@@ -1345,7 +1396,7 @@ def deduplicate_reads_by_coordinate(
         best_read_by: Quality criterion for selecting best read ('mapq' or 'avg_base_q')
         fragment_paired: Keep only fragments with both reads present
         fragment_mapped: Keep only fragments where both reads are mapped
-        
+
     Returns:
         Tuple of (deduplicated_reads, read_stats)
     """
@@ -1478,8 +1529,10 @@ def deduplicate_reads_by_coordinate(
                     fragment_group,
                     key=lambda f: calculate_fragment_quality_score(f, best_read_by),
                 )
-                cluster_name = best_fragment.get_cluster_name("coordinate", start_only, end_only)
-                
+                cluster_name = best_fragment.get_cluster_name(
+                    "coordinate", start_only, end_only
+                )
+
                 for fragment in fragment_group:
                     is_best = fragment == best_fragment
                     fragment.read1.set_tag("cn", cluster_name)
@@ -1487,7 +1540,7 @@ def deduplicate_reads_by_coordinate(
                     if not is_best:
                         fragment.read1.is_duplicate = True
                     deduplicated_reads.append(fragment.read1)
-                    
+
                     if fragment.read2:
                         fragment.read2.set_tag("cn", cluster_name)
                         fragment.read2.set_tag("cs", len(fragment_group))
@@ -1500,7 +1553,9 @@ def deduplicate_reads_by_coordinate(
                     fragment_group,
                     key=lambda f: calculate_fragment_quality_score(f, best_read_by),
                 )
-                cluster_name = best_fragment.get_cluster_name("coordinate", start_only, end_only)
+                cluster_name = best_fragment.get_cluster_name(
+                    "coordinate", start_only, end_only
+                )
                 best_fragment.read1.set_tag("cn", cluster_name)
                 best_fragment.read1.set_tag("cs", len(fragment_group))
                 deduplicated_reads.append(best_fragment.read1)
@@ -1518,8 +1573,6 @@ def deduplicate_reads_by_coordinate(
     return deduplicated_reads, stats
 
 
-
-
 def should_keep_fragment(fragment, fragment_paired, fragment_mapped):
     """Determine if a fragment should be kept based on fragment-level filtering options."""
     # fragment_paired: Keep only fragments with both reads present
@@ -1540,6 +1593,7 @@ def should_keep_fragment(fragment, fragment_paired, fragment_mapped):
 # MAIN PROCESSING FUNCTION
 # ============================================================================
 
+
 def process_bam(
     input_bam: str,
     output_bam: str,
@@ -1560,11 +1614,11 @@ def process_bam(
 ) -> bool:
     """
     Process BAM file using sequential reading and parallel window processing.
-    
+
     This function provides a unified interface for BAM deduplication with both
     UMI-based and coordinate-based methods. It includes comprehensive input
     validation, error handling, and progress reporting.
-    
+
     Args:
         input_bam: Path to input BAM file
         output_bam: Path to output BAM file
@@ -1582,10 +1636,10 @@ def process_bam(
         fragment_mapped: Keep only fragments where both reads are mapped (default: False)
         start_only: Use only start position for grouping (default: False)
         end_only: Use only end position for grouping (default: False)
-        
+
     Returns:
         bool: True if processing succeeded, False otherwise
-        
+
     Raises:
         ValueError: If input parameters are invalid
         FileNotFoundError: If input file doesn't exist
@@ -1598,29 +1652,32 @@ def process_bam(
     if not os.path.exists(input_bam):
         print(f"❌ Input file '{input_bam}' does not exist!")
         return False
-    
-    if not input_bam.endswith(('.bam', '.sam')):
+
+    if not input_bam.endswith((".bam", ".sam")):
         print(f"❌ Input file '{input_bam}' is not a BAM/SAM file!")
         return False
-    
-    if not output_bam.endswith(('.bam', '.sam')):
+
+    if not output_bam.endswith((".bam", ".sam")):
         print(f"❌ Output file '{output_bam}' must be a BAM/SAM file!")
         return False
 
     # Handle SAM files by converting to temporary BAM
     temp_bam_file = None
-    if input_bam.endswith('.sam'):
+    if input_bam.endswith(".sam"):
         import tempfile
+
         print("📝 Converting SAM to temporary BAM file...")
-        temp_bam_file = tempfile.mktemp(suffix='.bam')
-        
+        temp_bam_file = tempfile.mktemp(suffix=".bam")
+
         try:
             # Convert SAM to BAM
-            with pysam.AlignmentFile(input_bam, 'r') as sam_file:
-                with pysam.AlignmentFile(temp_bam_file, 'wb', header=sam_file.header) as bam_file:
+            with pysam.AlignmentFile(input_bam, "r") as sam_file:
+                with pysam.AlignmentFile(
+                    temp_bam_file, "wb", header=sam_file.header
+                ) as bam_file:
                     for read in sam_file:
                         bam_file.write(read)
-            
+
             # Create index for the temporary BAM file using threads
             pysam.index(temp_bam_file, "-@", str(max_processes or 1))
             input_bam = temp_bam_file  # Use the temporary BAM file for processing
@@ -1632,30 +1689,38 @@ def process_bam(
     # Validate method parameter
     valid_methods = ["umi", "coordinate"]
     if method not in valid_methods:
-        print(f"❌ Invalid method '{method}'. Valid methods are: {', '.join(valid_methods)}")
+        print(
+            f"❌ Invalid method '{method}'. Valid methods are: {', '.join(valid_methods)}"
+        )
         return False
-    
+
     # Validate quality criteria
     valid_quality_criteria = ["mapq", "avg_base_q"]
     if best_read_by not in valid_quality_criteria:
-        print(f"❌ Invalid quality criteria '{best_read_by}'. Valid criteria are: {', '.join(valid_quality_criteria)}")
+        print(
+            f"❌ Invalid quality criteria '{best_read_by}'. Valid criteria are: {', '.join(valid_quality_criteria)}"
+        )
         return False
-    
+
     # Validate numeric parameters
     if window_size <= 0:
         print(f"❌ Window size must be positive, got: {window_size}")
         return False
-    
+
     if max_pair_dist <= 0:
         print(f"❌ Max pair distance must be positive, got: {max_pair_dist}")
         return False
-    
+
     if not 0 <= min_edit_dist_frac <= 1:
-        print(f"❌ Min edit distance fraction must be between 0 and 1, got: {min_edit_dist_frac}")
+        print(
+            f"❌ Min edit distance fraction must be between 0 and 1, got: {min_edit_dist_frac}"
+        )
         return False
-    
+
     if not 0 <= min_frequency_ratio <= 1:
-        print(f"❌ Min frequency ratio must be between 0 and 1, got: {min_frequency_ratio}")
+        print(
+            f"❌ Min frequency ratio must be between 0 and 1, got: {min_frequency_ratio}"
+        )
         return False
 
     processor = None
@@ -1689,13 +1754,13 @@ def process_bam(
             start_only=start_only,
             end_only=end_only,
         )
-        
+
         # Set temp BAM file if it was created from SAM
         if temp_bam_file:
             processor.temp_bam_file = temp_bam_file
-        
+
         result = processor.process_bam()
-        
+
         return result
     except KeyboardInterrupt:
         print("\n⚠️  Processing interrupted by user (Ctrl+C)")
